/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  GenNoise [options] AFile

Purpose:
  Generate an audio file containing uncorrelated Gaussian noise samples

Description:
  This program generates zero-mean uncorrelated pseudo-random Gaussian
  deviates.  These white noise samples are written to an audio file.  Each
  invocation of this program generates a different random sequence.  Options
  are available to specify the number of samples, the sampling frequency and
  the data format of the output file.

Options:
  Output file name: AFile:
      Specifying "-" as the output file indicates that output is to be written
      to standard output.  If the output file type is not explicitly given (-F
      option), the extension of the output file name is used to determine the
      file type.
        ".au"   - AU audio file
        ".wav"  - RIFF WAVE file
        ".aif"  - AIFF-C audio file
        ".raw"  - Headerless file (native byte order)
        ".txt"  - Headerless file (text data)
  -d SDEV, --std_deviation=SDEV
      Standard deviation of the noise samples, default 1000.
  -x SEED, --seed=SEED
      Seed for the random number generator, default from current time.
  -n NSAMP, --number_frames=NSAMP
      Number of output samples to be generated.
  -s SFREQ, --srate=SFREQ
      Sampling frequency for the output audio file, default 8000.
  -F FTYPE, --file_type=FTYPE
      Output file type.
        "AU" or "au"             - AU audio file
        "WAVE" or "wave"         - RIFF WAVE file
        "AIFF-C" "aiff-c"        - AIFF-C audio file
        "noheader" or "noheader_native" - Headerless file (native byte
                                   order)
        "noheader_swap"          - Headerless file (byte swapped)
        "noheader_big-endian"    - Headerless file (big-endian byte
                                   order)
        "noheader_little-endian" - Headerless file (little-endian byte
                                   order)
  -D DFORMAT, --data_format=DFORMAT
      Data format for the output file, default integer16
        "mu-law8"   - 8-bit mu-law data
        "A-law8"    - 8-bit A-law data
        "unsigned8" - offset-binary 8-bit integer data
        "integer8"  - two's-complement 8-bit integer data
        "integer16" - two's-complement 16-bit integer data
        "integer24" - two's-complement 24-bit integer data
        "integer32" - two's-complement 32-bit integer data
        "float32"   - 32-bit floating-point data
        "float64"   - 64-bit floating-point data
        "text"      - text data
      The data formats available depend on the output file type.
      AU audio files:
        mu-law, A-law, 8-bit integer, 16-bit integer, 32-bit float, 64-bit
        float
      RIFF WAVE files:
        mu-law, A-law, offset-binary 8-bit integer, 16-bit integer
      AIFF-C audio files:
        mu-law, A-law, 8-bit integer, 16-bit integer
      Headerless files:
        all data formats
  -I INFO, --info=INFO
      Header information string.
  -h, --help
      Print a list of options and exit.
  -v, --version
      Print the version number and exit.

  By default, the audio file header for the output file contains a standard
  information string.
    Standard Header Information:
      date: 1994/01/25 19:19:39 UTC    date
      user: kabal@aldebaran            user
      program: GenNoise                program name
  This information can be changed with the header information string which is
  specified as one of the command line options.  Structured information records
  should adhere to the above format with a named field terminated by a colon,
  followed by numeric data or text.  Comments can follow as unstructured
  information.  For the purpose of this program, records are terminated by
  newline characters.  However in the header itself, the newline characters are
  replaced by nulls.  To place a newline character into the header, escape
  the newline character by preceding it with a '\' character.  If the first
  character of the user supplied header information string is a newline
  character, the header information string is appended to the standard header
  information.  If not, the user supplied header information string replaces
  the standard header information.

Author / version:
  P. Kabal / v4r0  1999/06/06  Copyright (C) 1999

-------------------------------------------------------------------------*/

static char rcsid[] = "$Id: GenNoise.c 1.46 1999/06/13 AFsp-v4r3 $";

#include <stdlib.h>	/* EXIT_SUCCESS */
#include <string.h>

#include <libtsp.h>
#include <libtsp/AFpar.h>
#include <AO.h>
#include "GenNoise.h"

#define MINV(a, b)	(((a) < (b)) ? (a) : (b))
#define MAXBUF	5120


int
main (int argc, const char *argv[])

{
  struct GN_FOpar FO;
  AFILE *AFp;
  FILE *fpinfo;
  int i, n, seed, Fformat;
  long int k;
  float rms, x[MAXBUF];

/* Get the input parameters */
  GNoptions (argc, argv, &rms, &seed, &FO);

/* If output is to stdout, use stderr for informational messages */
  if (strcmp (FO.Fname, "-") == 0)
    fpinfo = stderr;
  else
    fpinfo = stdout;

/* Open the output file */
  Fformat = AOsetFormat (&FO, NULL, 0);
  AFsetHinfo (FO.Hinfo);
  if (strcmp (FO.Fname, "-") != 0)
    FLbackup (FO.Fname);
  AFp = AFopenWrite (FO.Fname, Fformat, 1L, FO.Sfreq, fpinfo);

/* Generate the noise samples */
  MSrandSeed (seed);
  k = 0;
  while (k < FO.Nframe) {
    n = (int) MINV (FO.Nframe - k, MAXBUF);
    for (i = 0; i < n; ++i)
      x[i] = (float) MSfGaussRand ((double) rms);
    k += n;
    AFwriteData (AFp, x, n);
  }

/* Close the audio file */
  AFclose (AFp);

  return EXIT_SUCCESS;
}
