/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  void ADsetOPort (int audio_fd, const char Ports[NOPORT], double Vol,
                   double Balance, int Mute)

Purpose:
  Set the output port parameters for the Sun audio device

Description:
  This routine selects the output ports, sets the volume, sets the stereo
  balance, and controls muting for the Sun audio device.

Parameters:
   -> int audio_fd
      File descriptor for the audio device
   -> const char Ports[NOPORT]
      Output ports to be set/unset
   -> double Vol
      Output port volume, 0 to 100
   -> double Balance
      Output port stereo balance, 0 to 100
   -> int Mute
      Muting control
        -1  unmute, 0 no change, +1 mute

Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.17 $  $Date: 1999/06/14 16:20:51 $

-------------------------------------------------------------------------*/

static char rcsid[] = "$Id: ADsetOPort.c 1.17 1999/06/14 AFsp-v4r3 $";

#include <assert.h>
#include <stropts.h>		/* stream ioctl commands */
#include <unistd.h>		/* ioctl */

#include <libtsp.h>
#include <AD.h>

#define PGM		(UTgetProg ())


void
ADsetOPort (int audio_fd, const char Ports[NOPORT], double Vol, double Balance,
	    int Mute)

{
  audio_info_t aif;
  int iost, n, change;
  unsigned int port, portp;
  static const unsigned int port_mask[NOPORT] = {
    AUDIO_SPEAKER,
    AUDIO_HEADPHONE,
    AUDIO_LINE_OUT
  };

/* ==================== Select the output ports */
/* Read the play settings */
  iost = ioctl (audio_fd, AUDIO_GETINFO, &aif);
  if (iost == -1)
    UTerror ("%s: %s", PGM, ADM_GetErr);

/* Select the output ports */
  portp = aif.play.port;
  port = portp;
  for (n = 0; n < NOPORT; ++n) {
    if (Ports[n] == PORT_ADD)
      port = port | port_mask[n];
    else if (Ports[n] == PORT_DELETE)
      port = port & ~port_mask[n];
  }

/* Check against the available ports */
  if ((port & ~aif.play.avail_ports) != 0)
    UThalt ("%s: %s", PGM, ADM_PortErr);

/* Initialize the audio device structure */
  AUDIO_INITINFO (&aif);
  change = 0;

/* Select the ports */
  if (port != portp) {
    aif.play.port = port;
    change = 1;
  }

/* Volume setting */
  if (Vol != VOL_UNDEF) {
    assert (Vol >= 0.0 && Vol <= 100.0);
    aif.play.gain = (unsigned int) ((Vol * AUDIO_MAX_GAIN / 100.0) + 0.5);
    change = 1;
  }

/* Stereo balance setting */
  if (Balance != BALANCE_UNDEF) {
    assert (Balance >= 0.0 && Balance <= 100.0);
    aif.play.balance = (Balance * AUDIO_RIGHT_BALANCE / 100.0) + 0.5;
    change = 1;
  }

/* Mute setting */
  if (Mute != MUTE_DEFAULT) {
    aif.output_muted = (Mute > 0) ? 1 : 0;
    change = 1;
  }

/* Update the audio device settings if they have changed */
/* Note:
   Under Solaris, the AUDIO_SETINFO ioctl always returns an error code (even
   for an ioctl immediately following AUDIO_INITINFO).  Example source files
   in /usr/demo/SOUNDS ignore the return code.
*/
  if (change) {
    iost = ioctl (audio_fd, AUDIO_SETINFO, &aif);
#ifndef SOLARIS
    if (iost == -1)
      UTerror ("%s: %s", PGM, ADM_SetErr);
#endif
  }

  return;
}
