/*-------------- Telecommunications & Signal Processing Lab ---------------

Routine:
  AFILE *AFrdWVhead (FILE *fp)

Purpose:
  Get file format information from a RIFF WAVE file

Description:
  This routine reads the header for a RIFF WAVE file.  The header information
  is used to set the file data format information in the audio file pointer
  structure.

  RIFF WAVE file:
   Offset Length Type    Contents
      0     4    char   File identifier ("RIFF")
      4     4    int    RIFF chunk length
      8     4    char   Form identifier ("WAVE")
    ...   ...    ...    ...
      f     4    char   Format chunk identifier ("fmt ")
    f+4     4    int    Format chunk length
    f+8     2    int    Audio data type
    f+10    2    int    Number of interleaved channels
    f+12    4    int    Sample rate
    f+16    4    int    Average bytes/sec
    f+20    2    int    Block align
    f+22    2    int    Data word length (bits)
    ...    ...   ...    ...
      d     4    char   Data chunk identifier ("data")
    d+4     4    int    Data chunk length
    d+8    ...   ...    Audio data

Parameters:
  <-  AFILE *AFrdWVhead
      Audio file pointer for the audio file
   -> FILE *fp
      File pointer for the file

Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.65 $  $Date: 1999/06/11 22:24:39 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFrdWVhead.c 1.65 1999/06/11 AFsp-v4r3 $";

#include <setjmp.h>
#include <string.h>

#include <libtsp.h>
#include <libtsp/AFdataio.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#include <libtsp/AFpar.h>
#include <libtsp/WVpar.h>
#include <libtsp/nucleus.h>

#define ICEILV(n,m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */
#define NELEM(array)	((sizeof array) / (sizeof array[0]))
#define RNDUPV(n,m)	((m) * ICEILV (n, m))		/* Round up */

#define SAME_CSTR(str,ref) 	(memcmp (str, ref, sizeof (str)) == 0)

#define ALIGN		2	/* Chunks padded out to a multiple of ALIGN */

#define WV_FMT_CKSIZE	16
#define WV_LHMIN	(8 + 4 + 8 + WV_FMT_CKSIZE + 8)

/* setjmp / longjmp environment */
extern jmp_buf AFR_JMPENV;

static void
AF_decFMT (struct WV_PCMfmt *CkPCM, struct AF_read *AFr);
static int
AF_rdDISP_text (FILE *fp, int Size, struct AF_info *Hinfo);
static int
AF_rdFMT (FILE *fp, struct WV_PCMfmt *CkPCM);
static int
AF_rdLIST_INFO (FILE *fp, int Size, struct AF_info *Hinfo);
static long int
AF_rdRIFF_WAVE (FILE *fp);
static void
AF_UnsFormat (int FormatTag);


AFILE *
AFrdWVhead (FILE *fp)

{
  AFILE *AFp;
  int AtData, FoundAFsp;
  long int offs, LRIFF, Dstart;
  char Info[AF_MAXINFO];
  struct WV_CkPreamb CkHead;
  struct WV_PCMfmt CkPCM;
  struct AF_read AFr;

/* Set the long jump environment; on error return a NULL */
  if (setjmp (AFR_JMPENV))
    return NULL;	/* Return from a header read error */

/* Defaults and initial values */
  AFr.DFormat.Format = FD_UNDEF;
  AFr.NData.Nsamp = AF_NSAMP_UNDEF;
  AFr.NData.Ldata = -1L;
  AFr.Hinfo.Info = Info;
  AFr.Hinfo.N = 0;
  AFr.Hinfo.Nmax = AF_MAXINFO;

/* Check the file magic for a RIFF/WAVE file */
  LRIFF = AF_rdRIFF_WAVE (fp);
  offs = 12L;	/* Positioned after RIFF/WAVE preamble */

  Dstart = 0L;
  AtData = 0;
  FoundAFsp = 0;
  while (offs < LRIFF) {

    /* Read the chunk preamble */
    offs += RHEAD_S (fp, CkHead.ckid);
    offs += RHEAD_V (fp, CkHead.cksize, DS_EL);

    /* fmt chunk */
    if (SAME_CSTR (CkHead.ckid, "fmt ")) {
      CkPCM.cksize = CkHead.cksize;
      offs += AF_rdFMT (fp, &CkPCM);
      AF_decFMT (&CkPCM, &AFr);
    }

    /* data chunk */
    else if (SAME_CSTR (CkHead.ckid, "data")) {
      AFr.NData.Ldata = CkHead.cksize;
      Dstart = offs;
      if (RNDUPV (CkHead.cksize + offs, ALIGN) >= (unsigned int) LRIFF ||
	  ! FLseekable (fp)) {
	AtData = 1;
	break;
      }
      else {
	offs += RSKIP (fp, RNDUPV (CkHead.cksize, ALIGN));
	AtData = 0;
      }
    }

    /* Text chunks */
    else if (SAME_CSTR (CkHead.ckid, "afsp")) {
      FoundAFsp = 1;
      offs += AFrdHinfo (fp, (int) CkHead.cksize, &AFr.Hinfo, ALIGN);
    }
    else if (SAME_CSTR (CkHead.ckid, "DISP") && ! FoundAFsp)
      offs += AF_rdDISP_text (fp, (int) CkHead.cksize, &AFr.Hinfo);
    else if (SAME_CSTR (CkHead.ckid, "LIST") && ! FoundAFsp)
      offs += AF_rdLIST_INFO (fp, (int) CkHead.cksize, &AFr.Hinfo);

    /* Miscellaneous chunks */
    else
      offs += RSKIP (fp, RNDUPV (CkHead.cksize, ALIGN));
  }

  if (AFr.DFormat.Format == FD_UNDEF || AFr.NData.Ldata < 0L) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadHead);
    return NULL;
  }
  if (! AtData && offs != LRIFF)
    UTwarn ("AFrdWVhead - %s", AFM_WV_FixRIFF);

  /* Position at the start of data */
  if (! AtData) {
    if (AFseek (fp, Dstart, NULL))
      return NULL;
  }

/* Set the parameters for file access */
  AFp = AFsetRead (fp, FT_WAVE, &AFr, AF_FIX_LDATA_HIGH);

  return AFp;
}

/* Check the RIFF/WAVE file preamble:  On error return via longjmp */


static long int
AF_rdRIFF_WAVE (FILE *fp)

{
  int offs;
  long int LRIFF, Lfile;
  struct WV_CkPreamb CkHead;

  offs = RHEAD_S (fp, CkHead.ckid);
  if (! SAME_CSTR (CkHead.ckid, FM_RIFF)) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadId);
    longjmp (AFR_JMPENV, 1);
  }

  offs += RHEAD_V (fp, CkHead.cksize, DS_EL);
  LRIFF = CkHead.cksize + offs;
  if (LRIFF < WV_LHMIN) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadRIFF);
    longjmp (AFR_JMPENV, 1);
  }

  if (FLseekable (fp)) {
    Lfile = FLfileSize (fp);
    if (LRIFF > Lfile) {
      LRIFF = Lfile;
      UTwarn ("AFrdWVhead - %s", AFM_WV_FixRIFF);
    }
  }
 
  offs += RHEAD_S (fp, CkHead.ckid);
  if (! SAME_CSTR (CkHead.ckid, FM_WAVE)) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadId);
    longjmp (AFR_JMPENV, 1);
  }

  return LRIFF;
}

/* Read the format chunk */


static int
AF_rdFMT (FILE *fp, struct WV_PCMfmt *CkPCM)

{
  int offs;

  if (CkPCM->cksize < WV_FMT_CKSIZE) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadPCM);
    longjmp (AFR_JMPENV, 1);
  }

  offs  = RHEAD_V (fp, CkPCM->FormatTag, DS_EL);
  offs += RHEAD_V (fp, CkPCM->Channels, DS_EL);
  offs += RHEAD_V (fp, CkPCM->SamplesPerSec, DS_EL);
  offs += RHEAD_V (fp, CkPCM->AvgBytesPerSec, DS_EL);
  offs += RHEAD_V (fp, CkPCM->BlockAlign, DS_EL);
  offs += RHEAD_V (fp, CkPCM->BitsPerSample, DS_EL);

  /* Skip over any extra data at the end of the fmt chunk */
  if (offs > (int) CkPCM->cksize)
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadPCM);
  else
    offs += RSKIP (fp, RNDUPV (CkPCM->cksize, ALIGN) - offs);

  return offs;
}

/* Decode the data format: On error return via longjmp */


static void
AF_decFMT (struct WV_PCMfmt *CkPCM, struct AF_read *AFr)

{
  switch (CkPCM->FormatTag) {
  case WAVE_FORMAT_PCM:
    if (CkPCM->BitsPerSample > 8 && CkPCM->BitsPerSample <= 16) {
      AFr->DFormat.Format = FD_INT16;
      AFr->DFormat.ScaleF = WV_SF_PCM16;
    }
    /* Special case: IEEE float from CoolEdit (flagged as 24-bit integer) */
    else if (CkPCM->BitsPerSample == 24 && 
	     CkPCM->AvgBytesPerSec == 4 * CkPCM->Channels *
                                      CkPCM->SamplesPerSec) {
      AFr->DFormat.Format = FD_FLOAT32;
      AFr->DFormat.ScaleF = WV_SF_FLOAT32_X;
      CkPCM->BitsPerSample = 8 * FDL_FLOAT32;
      CkPCM->FormatTag = WAVE_FORMAT_IEEE_FLOAT;
      if (! UTcheckIEEE ())
	UTwarn ("AFrdAIhead - %s", AFM_WV_NoIEEE);
    }
    else if (CkPCM->BitsPerSample > 16 && CkPCM->BitsPerSample <= 24) {
      AFr->DFormat.Format = FD_INT24;
      AFr->DFormat.ScaleF = WV_SF_PCM24;
    }
    else if (CkPCM->BitsPerSample > 24 && CkPCM->BitsPerSample <= 32) {
      AFr->DFormat.Format = FD_INT32;
      AFr->DFormat.ScaleF = WV_SF_PCM32;
    }
    else if (CkPCM->BitsPerSample > 0 && CkPCM->BitsPerSample <= 8) {
      AFr->DFormat.Format = FD_UINT8;
      AFr->DFormat.ScaleF = WV_SF_PCM8;
    }
    else {
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_UnsDSize,
	      (int) CkPCM->BitsPerSample);
      longjmp (AFR_JMPENV, 1);
    }
    break;
  case WAVE_FORMAT_MULAW:
    if (CkPCM->BitsPerSample != 8)
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_BadMulaw,
	      (int) CkPCM->BitsPerSample);
    AFr->DFormat.Format = FD_MULAW8;
    AFr->DFormat.ScaleF = WV_SF_MULAW;
    break;
  case WAVE_FORMAT_ALAW:
    if (CkPCM->BitsPerSample != 8)
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_BadAlaw,
	      (int) CkPCM->BitsPerSample);
    AFr->DFormat.Format = FD_ALAW8;
    AFr->DFormat.ScaleF = WV_SF_ALAW;
    break;
  case WAVE_FORMAT_IEEE_FLOAT:
    if (CkPCM->BitsPerSample == 8 * FDL_FLOAT32) {
      AFr->DFormat.Format = FD_FLOAT32;
      AFr->DFormat.ScaleF = WV_SF_FLOAT32;
    }
    else if (CkPCM->BitsPerSample == 8 * FDL_FLOAT64) {
      AFr->DFormat.Format = FD_FLOAT64;
      AFr->DFormat.ScaleF = WV_SF_FLOAT64;
    }
    else
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_BadFloat,
	      (int) CkPCM->BitsPerSample);
    if (! UTcheckIEEE ())
      UTwarn ("AFrdWVhead - %s", AFM_WV_NoIEEE);
    break;

  /* Unsupported data formats */
  default:
    AF_UnsFormat ((int) CkPCM->FormatTag);
    longjmp (AFR_JMPENV, 1);
  }

  /* Consistency check */
  if (CkPCM->AvgBytesPerSec != CkPCM->Channels * CkPCM->SamplesPerSec *
      ICEILV (CkPCM->BitsPerSample, 8))
    UTwarn ("AFrdWVhead - %s", AFM_WV_MisNBytes);

  AFr->DFormat.Swapb = DS_EL;

  AFr->Sfreq = (double) CkPCM->SamplesPerSec;
  AFr->NData.Nchan = (long int) CkPCM->Channels;

  return;
}

/* Unsupported data format */

static void
AF_UnsFormat (int FormatTag)

{
  switch (FormatTag) {
  case WAVE_FORMAT_ADPCM:
    UTwarn ("AFrdWVhead - %s: \"%s\"", AFM_WV_UnsData, "G.721 ADPCM");
    break;
  case WAVE_FORMAT_MPEGLAYER3:
    UTwarn ("AFrdWVhead - %s: \"%s\"", AFM_WV_UnsData, "MPEG-1 Layer 3");
    break;
  case WAVE_FORMAT_MSG723:
    UTwarn ("AFrdWVhead - %s: \"%s\"", AFM_WV_UnsData, "G.723.1 coding");
    break;
  default:
    UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_UnsData, FormatTag);
    break;
  }
}

/* Read a DISP chunk */


static int
AF_rdDISP_text (FILE *fp, int Size, struct AF_info *Hinfo)

{
  int offs;
  uint4_t DISP_ID;

  offs = RHEAD_V (fp, DISP_ID, DS_EL);
  if (DISP_ID == CF_TEXT)
    offs += AFrdHtext (fp, Size - offs, "display_name: ", Hinfo, ALIGN);

  offs += RSKIP (fp, RNDUPV (Size, ALIGN) - offs);

  return offs;
}     

/* Read the LIST-INFO records from the header */


struct WV_LI{
  char ckid[4];
  char *key;
};
static const struct WV_LI IID[] = {
  {"IARL", "archival_location: "},
  {"IART", "artist: "},
  {"ICMS", "commissioned: "},
  {"ICMT", "comments: "},
  {"ICOP", "copyright: "},
  {"ICRD", "creation_date: "},
  {"IENG", "engineer: "},
  {"IGNR", "genre: "},
  {"IKEY", "keywords: "},
  {"IMED", "medium: "},
  {"INAM", "name: "},
  {"IPRD", "product: "},
  {"ISBJ", "subject: "},
  {"ISFT", "software: "},
  {"ISRC", "source: "},
  {"ISRF", "source_form: "},
  {"ITCH", "technician: "}
};

#define NIID	NELEM (IID)


static int
AF_rdLIST_INFO (FILE *fp, int Size, struct AF_info *Hinfo)

{
  int i, offs;
  char ID[4], key[7];
  struct WV_CkPreamb CkHead;

  offs = RHEAD_S (fp, ID);
  if (SAME_CSTR (ID, "INFO")) {

    while (offs < Size) {
      offs += RHEAD_S (fp, CkHead.ckid);
      offs += RHEAD_V (fp, CkHead.cksize, DS_EL);

      /* Look for standard INFO ID values */
      for (i = 0; i < (int) NIID; ++i) {
	if (SAME_CSTR (CkHead.ckid, IID[i].ckid)) {
	  offs += AFrdHtext (fp, (int) CkHead.cksize, IID[i].key,
			     Hinfo, ALIGN);
	  break;
	}
      }

      /* No match, use the INFO ID field as the information record keyword */
      if (i == NIID) {
	strncpy (key, CkHead.ckid, 4);
	strcpy (&key[4], ": ");
	offs += AFrdHtext (fp, (int) CkHead.cksize, key, Hinfo, ALIGN);
      }
    }
  }
  offs += RSKIP (fp, RNDUPV (Size, ALIGN) - offs);

  return offs;
}
